/*
   name:    CircularQueue.rex
   author:  Rony G. Flatscher, Rick McGuire
   date:    2006-05-09, 2006-05-11, 2006-05-15, 2006-05-16, 2006-05-17, 2006-05-27
   purpose: allow to use a circular queue (buffer) of predefined size; if queue is full
            new entries overwrite existing ones in a circular fashion; also allows for
            resizing the circular queue at runtime
   license: CPL 1.0

   documentation:

   The CircularQueue class
   =======================

      This is a collection class which allows for storing objects in a circular queue
      of a predefined size. Once the end of the queue has been reached, new item
      objects are inserted from the beginning, replacing earlier entries.

      The collected objects can be processed in FIFO (first in, first out) or in a
      stack-like LIFO (last in, first out) order.



   OF (Class Method)
   =================

          +-,----+
          v      |
   >>-OF(---item-+-)--------------------------<<

   Returns a newly created CircularQueue object containing the specified item
   objects in the order specified. The number of listed items determines the
   size of the circular queue.


   INIT
   ====

   >>-INIT(size)----------------------------------<<

   Constructor method invoked by the NEW method, which determines the size of
   the circular queue.


   MAKEARRAY
   =========
                 +-Fifo-----+
   >>-MAKEARRAY(-+----------+-)----------------------<<
                 +-,-order--+

   Returns a single-index array containing the items of the circular queue in the
   specified order.

   The following orders can be used. (Only the capitalized letter is needed; all
   characters following it are ignored.)

   Fifo
      "first-in, first-out", default

   Lifo
      stacklike, "last-in, first-out"


   PUSH
   ====

   >>-PUSH(item)----------------------------------<<

   Makes the object item a member item of the circular queue, inserting the item
   object in front of the first item in the queue. The pushed item object will be
   the new first item in the circular queue.

   If the circular queue is full, than the last item stored in the circular queue
   will be deleted, before the insertion takes place. In this case the deleted item
   will be returned, otherwise .nil.



   QUEUE
   =====

   >>-QUEUE(item)---------------------------------<<

   Makes the object item a member item of the circular queue, inserting the item
   at the end of the circular queue.

   If the circular queue is full, than the first item will be deleted, before the
   insertion takes place. In this case the deleted item will be returned, otherwise .nil.



   RESIZE
   ======

                      +-Fifo-----+
   >>-RESIZE(-newSize-+----------+-)----------------------<<
                      +-,-order--+

   Resizes the circular queue object to be able to contain "newSize" items.
   If the previous size was larger than "newSize", than the now superfluous
   items are removed in the specified order.

   The following orders can be used. (Only the capitalized letter is needed; all
   characters following it are ignored.)

   Fifo
      "first-in, first-out", default: keeps the newest entries

   Lifo
      stacklike, "last-in, first-out": keeps the oldest entries

   Note: resizing with a value of "0" effectively removes all items from the
   circular queue.


   SIZE
   =====

   >>-SIZE----------------------------------------<<

   Returns the maximum number of objects that can be stored in the circular queue.



   SUPPLIER
   =========
                +-Fifo-----+
   >>-SUPPLIER(-+----------+-)----------------------<<
                +-,-order--+

   Returns a supplier object for the collection. After you have obtained a supplier,
   you can send it messages (see The Supplier Class) to enumerate all the items
   that were in the circular queue at the time of the supplier's creation.

   The supplier will enumerate the items in the specified order. (Only the capitalized
   letter is needed; all characters following it are ignored.)

   Fifo
      "first-in, first-out", default

   Lifo
      stacklike, "last-in, first-out"


   STRING
   ======
              +-----","---+  +-Fifo----+
   >>-STRING(-+-----------+--+---------+-)---------<<
              +-delimiter-+  +-,-order-+

   Returns a string object that concatenates the string values of the collected
   item objects, using the delimiter string to delimit them, in the specified order.

   If the delimiter string argument is omitted the comma character is used as the
   default delimiter string.

   The following orders can be used. (Only the capitalized letter is needed;
   all characters following it are ignored.)

   Fifo
      "first-in, first-out", default

   Lifo
      stacklike, "last-in, first-out"

   --------------- CircularQueue example --------------

      -- create a circular buffer with five items
   u=.circularQueue~of("a", "b", "c", "d", "e")
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say

   u~resize(4, "FIFO")     -- resize fifo-style (keep newest)
   say "after resizing to 4 items in FIFO style (keeping the newest):"
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say

   u~resize(2, "LILO")     -- resize lifo-style (keep oldest)
   say "after resizing to 2 items in LIFO style (keeping the oldest):"
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say

   u~resize(0)             -- resize lifo-style (keep oldest)
   say "after resizing to 0 items, thereby deleting all items:"
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say

   u~resize(2)             -- resize lifo-style (keep oldest)
   say "after resizing to 2, size="u~size "and items="u~items
   u~~queue('x')~~queue('y')~~queue('z')
   say "after queuing the three items 'x', 'y', 'z':"
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say

   u~~push('1')~~push('2')~~push('3')
   say "after pushing the three items '1', '2', '3':"
   say "content: ["u"]," "content (LIFO): ["u~string("->","L")"]"
   say


/*
Output:

   content: [a,b,c,d,e], content (LIFO): [e->d->c->b->a]

   after resizing to 4 items in FIFO style (keeping the newest):
   content: [b,c,d,e], content (LIFO): [e->d->c->b]

   after resizing to 2 items in LIFO style (keeping the oldest):
   content: [b,c], content (LIFO): [c->b]

   after resizing to 0 items, thereby deleting all items:
   content: [], content (LIFO): []

   after resizing to 2, size=2 and items=0
   after queuing the three items 'x', 'y', 'z':
   content: [y,z], content (LIFO): [z->y]

   after pushing the three items '1', '2', '3':
   content: [3,2], content (LIFO): [2->3]

*/


*/

   -- circular buffer (round-robin queue) of a defined size
::class "CircularQueue" subclass queue public


::method of class

  signal on syntax
  storage=self~new(arg())     -- create a circular queue of appropriate size

  do i=1 to arg()
     storage~queue(arg(i))    -- place item object into the circular queue
  end
  return storage              -- return instance

syntax: raise propagate


::method init                 -- create an instance, memorizes the size
  expose size
  parse arg size

  signal on syntax
  if arg(1, "O") then
     raise syntax 93.903 array (1)  -- raise an error
  if (arg() > 1) then               -- too many arguments given?
     raise syntax 93.902 array (1)  -- raise an error

  -- check for whole number >= 0, raise error
  if datatype(size, "Whole")=.false | size<0  then
     raise syntax 93.906 array ("size", size)

  self~init:super
  return

syntax: raise propagate



::METHOD size                 -- return size
  expose size

  signal on syntax
  if (arg() > 0) then               -- too many arguments given?
     raise syntax 93.902 array (0)  -- raise an error
 return size

syntax: raise propagate


::METHOD resize               -- allow to resize the circularQueue
  expose size

  signal on syntax
  if arg(1, "O") then
     raise syntax 93.903 array (1)  -- raise an error
  if (arg() > 2) then               -- too many arguments given?
     raise syntax 93.902 array (2)  -- raise an error

   parse upper arg newSize, order +1   -- parse arguments

   -- check for whole number >= 0, raise error
   if datatype(newSize, "Whole")=.false | newSize<0  then
      raise syntax 93.906 array (1, arg(1))

   if arg(2,"Exists") then
   do
      if pos(order, "FL")=0 then  -- oops: not a valid argument!
         raise syntax 93.914 array (2, '"F" (FIFO) or "L" (LIFO)', arg(2))
   end


   if newSize<size then       -- too many items, we need to remove the superfluous ones
   do
      removeNr=size-newSize   -- number of elements to remove

      if order="L" then       -- LIFO: remove the newest entries (keep the oldest)
         idx=self~items-removeNr+1    -- determine the index to which (includingly) the removes need to take place
      else                    -- FIFO: remove the oldest entries (keep the newest)
         idx=1                -- remove the first element(s)

      do removeNr             -- repeat removal
         self~remove(idx)     -- remove item
      end
   end
   size=newSize               -- remember new size
   return newSize             -- return the new size

syntax: raise propagate



::METHOD queue                -- put an element into the queue, possibly replacing an existing one
  expose size

  signal on syntax
  if arg(1, "O") then
     raise syntax 93.903 array (1)  -- raise an error
  if (arg() > 1) then               -- too many arguments given?
     raise syntax 93.902 array (1)  -- raise an error

  if size=0 then return .nil  -- size is 0, do not process!

  if self~items >= size then
     item=self~pull           -- if we've gone over the size, remove this
  else
     item=.nil                -- return value a .nil to indicate that we were not forced to remove an existing item

  forward class (super) continue -- allow the superclass to handle adding the item
  return item                 -- return pulled item

syntax: raise propagate


::METHOD push
  expose size

  signal on syntax
  if arg(1, "O") then
     raise syntax 93.903 array (1)  -- raise an error
  if (arg() > 1) then               -- too many arguments given?
     raise syntax 93.902 array (1)  -- raise an error

  if size=0 then return .nil  -- size is 0, do not process !

  if self~items >= size then
     item=self~remove(self~items)   -- remove the end item.
  else
     item=.nil                -- return value a .nil to indicate that we were not forced to remove an existing item

  forward class (super) continue    -- allow the superclass to handle adding the item
  return item                 -- return removed item

syntax: raise propagate


-----------------------------


::method makearray            -- return a supplier object
  expose size storage

  signal on syntax
  if (arg() > 1) then               -- too many arguments given?
     raise syntax 93.902 array (1)  -- raise an error

  parse upper arg order +1

  if arg(1,"Exists") then
  do
     if pos(order, "FL")=0 then  -- oops: not a valid argument!
        raise syntax 93.914 array (1, '"F" (FIFO) or "L" (LIFO)', arg(1))
  end

  if order="L" then           -- in LIFO (last-in, first-out) order (stack-like)
  do
     aCopy=self~copy          -- work on a copy (= snapshot)
     items=aCopy~items        -- get number of items
     valArr=.array~new(items)
     k=0
     do i=items to 1 by -1
        k=k+1
        valArr[k]=aCopy~at(i)
     end
     return valArr
  end
  else
  do
     arr=.array~new(0)
     FORWARD class (super) arguments (arr)-- forward to superclass, remove possible argument
     -- FORWARD class super array ()-- forward to superclass, remove possible argument
  end

syntax: raise propagate


::method supplier             -- return a supplier object
  expose size

  signal on syntax
  if (arg() > 1) then               -- too many arguments given?
     raise syntax 93.902 array (1)  -- raise an error

  parse upper arg order +1

  if arg(1,"Exists") then
  do
     if pos(order, "FL")=0 then  -- oops: not a valid argument!
        raise syntax 93.914 array (1, '"F" (FIFO) or "L" (LIFO)', arg(1))
  end

  if order="L" then           -- in LIFO (last-in, first-out) order (stack-like)
  do
     aCopy=self~copy          -- work on a copy (= snapshot)
     items=aCopy~items        -- get number of items
     valArr=.array~new(items) -- create value (item) array
     idxArr=.array~new(items) -- create index array
     k=0
     do i=items to 1 by -1    -- iterate through circular queue
        k=k+1
        valArr[k]=aCopy~at(i)
        idxArr[k]=i
     end
     return .supplier~new(valArr, idxArr)
  end
  else                        -- let the superclass create and return the supplier object
  do
     FORWARD class (super) array ()-- forward to superclass, remove possible argument
  end

syntax: raise propagate


::method string               -- create a string representation, using the optional delimiter
  expose size

  signal on syntax
  if (arg() > 2) then               -- too many arguments given?
     raise syntax 93.902 array (2)  -- raise an error

  if self~items=0 then        -- return empty strings, if no elements
     return ""

  parse arg delimiter, order +1  -- do not translate to uppercase as to preserve the case of "delimiter"

  if arg(2,"Exists") then
  do
     order=translate(order)   -- make sure it is translated to uppercase
     if pos(order, "FL")=0 then  -- oops: not a valid argument!
        raise syntax 93.914 array (2, '"F" (FIFO) or "L" (LIFO)', arg(2))
  end

  if arg(1,"Omitted") then    -- default to comma as a separator, if argument was omitted
     delimiter=","

  mb=.MutableBuffer~new       -- use a MutableBuffer for the string chunks
                              -- use MAKEARRAY for a snapshot in the desired order
  if arg(2, "Omitted") then   -- order was omitted
     array=self~makearray     -- omit order as well
  else
     array=self~makearray(order) -- use supplied order

  k=.false                    -- append supplied delimiter?
  do i=1 to array~items
     if k then                -- at least one item already there, hence use delimiter
        mb~append(delimiter)
     else                     -- first iteration
        k=.true

     mb~append(array[i]~string) -- get element from array, get its string value
  end

  return mb~string            -- now turn MutableBuffer object into a plain string

syntax: raise propagate

